package anywheresoftware.b4a.objects;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.util.Map.Entry;

import jcifs.Config;
import jcifs.smb.NtlmPasswordAuthentication;
import jcifs.smb.SmbException;
import jcifs.smb.SmbFile;
import anywheresoftware.b4a.AbsObjectWrapper;
import anywheresoftware.b4a.BA;
import anywheresoftware.b4a.BA.DependsOn;
import anywheresoftware.b4a.BA.Events;
import anywheresoftware.b4a.BA.Permissions;
import anywheresoftware.b4a.BA.ShortName;
import anywheresoftware.b4a.BA.Version;
import anywheresoftware.b4a.objects.collections.Map;
import anywheresoftware.b4a.objects.streams.File;
import anywheresoftware.b4a.objects.streams.File.InputStreamWrapper;
import anywheresoftware.b4a.objects.streams.File.OutputStreamWrapper;

/**
 *This library is based on <link>JCIFS|http://jcifs.samba.org/</link> library.
 *It provides access to Microsoft Windows network file system.
 *See this <link>tutorial|http://www.basic4ppc.com/forum/basic4android-getting-started-tutorials/17179-android-smb-cifs-tutorial.html</link>.
 */
@Permissions(values = {"android.permission.INTERNET"})
@Events(values ={"DownloadCompleted (Url As String, RemoteFile As String, Success As Boolean)",
		"UploadCompleted (Url As String, RemoteFile As String, Success As Boolean)",
		"DeleteCompleted (Url As String, RemoteFile As String, Success As Boolean)",
		"ListCompleted (Url As String, Success As Boolean, Entries() As SMBFile)",
		"MakeDirCompleted (Url As String, RemoteFile As String, Success As Boolean)",
		"RenameCompleted (Url As String, DestRemoteFile As String, Success As Boolean)",
		"CopyCompleted (DestUrl As String, DestRemoteFile As String, Success As Boolean)",
		"FreeSpaceCompleted (Url As String, RemoteFile As String, Size As Long, Success As Boolean)"
		})
@Version(1.1f)
@DependsOn(values={"jcifs"})
@ShortName("SMB")
public class SMB
{
	
	private static final double version = 1.1;
	//		added SMB getDiskFreeSpace, MakeDirs, CopyFile and RenameFile
	//		added SmbFileWrapper CanRead, CanWrite and IsHidden

	/**
	 *Returns the version of the library.
	 */
	public double getVersion()
	{
		return version;
	}
	/**
	 *This library is based on <link>JCIFS|http://jcifs.samba.org/</link> library.
	 *It provides access to Microsoft Windows network file system.
	 *See this <link>tutorial|http://www.basic4ppc.com/forum/basic4android-getting-started-tutorials/17179-android-smb-cifs-tutorial.html</link>.
	 *
	 *JCIFS and therefore also this library is licensed under the GNU Lesser General Public License.
	 *<link>http://www.gnu.org/licenses/lgpl.html|http://www.gnu.org/licenses/lgpl.html</link>
	 *A copy of the Lesser General Public License is in the provided archive.
	 *
	 *Copyright 2013 Anywhere Software
	 */
	public void LIBRARY_DOC()
	{
	}


	private String eventName;
	private String user = "", password = "", domain = "";
	
	/**
	 *Initializes the object.
	 */
	public void Initialize(String EventName) throws IOException {
		Initialize2(EventName, null);
	}

	/**
	 *Sets the credentials that should be used to make the connection.
	 */
	public void SetCredentials(String Username, String Password, String Domain) {
		user = Username;
		password = Password;
		domain = Domain;
	}

	private SmbFile createSmb(String Url, String RemoteFile) throws MalformedURLException {
		if (user.length() > 0 || password.length() > 0)
			return new SmbFile(Url, RemoteFile, new NtlmPasswordAuthentication(domain, user, password));
		else
			return new SmbFile(Url, RemoteFile);
	}

	/**
	 *Initializes the object and sets the specified map of properties as explained on this page:
	 *<link>SMB properties|http://jcifs.samba.org/src/docs/api/overview-summary.html</link>
	 */
	public void Initialize2(String EventName, Map Properties) throws IOException {
		Config.load(null); // Erel missed this out I think
		this.eventName = EventName.toLowerCase(BA.cul); 
		if (Properties != null) {
			for(Entry<Object, Object> e : Properties.getObject().entrySet()) {
				Config.setProperty(String.valueOf(e.getKey()), String.valueOf(e.getValue()));
			}
		}
	}

	/**
	 *This method will delete the file or directory specified by this SmbFile.
	 *If the target is a directory, the contents of the directory will be deleted as well.
	 *If a file within the directory or it's sub-directories is marked read-only, the read-only status will be removed and the file will be deleted. 
	 *The DeleteCompleted event will be raised when the process completes.
	 *Example: <code>SMB1.DeleteFile("smb://USER-PC/", "1.jpg")</code>
	 */
	public void DeleteFile(final BA ba, final String Url, final String RemoteFile) {
		Runnable r = new Runnable() {
			@Override
			public void run() {
				try {
					SmbFile sf = createSmb(Url, RemoteFile);
					if (sf.exists())
						sf.delete();
					ba.raiseEventFromDifferentThread(null, null, 0, eventName + "_deletecompleted", false, new Object[] { Url, RemoteFile, true});
				}
				catch (Exception e) {
					ba.setLastException(e);
					ba.raiseEventFromDifferentThread(null, null, 0, eventName + "_deletecompleted", false, new Object[] { Url, RemoteFile, false});
				}
			}
		};
		ba.submitRunnable(r, this, 0);
	}

	/**
	 *Uploads a local file.
	 *The UploadCompleted event will be raised when the process completes.
	 *Example: <code>SMB1.UploadFile(File.DirRootExternal, "1.jpg", "smb://USER-PC/", "1.jpg")</code>
	 *
	 */
	public void UploadFile(final BA ba, String LocalDir, String LocalFile, final String Url, final String RemoteFile) throws IOException {
		InputStreamWrapper in = new InputStreamWrapper();
		in = File.OpenInput(LocalDir, LocalFile);
		UploadFile2(ba, in.getObject(), Url, RemoteFile);
	}

	/**
	 *Similar to UploadFile. Expects an InputStream instead of the local file.
	 */
	public void UploadFile2(final BA ba, final InputStream In, final String Url, final String RemoteFile) {
		Runnable r = new Runnable() {
			@Override
			public void run() {
				try {
					SmbFile sf = createSmb(Url, RemoteFile);
					OutputStream out = sf.getOutputStream();
					File.Copy2(In, out);
					out.close();
					ba.raiseEventFromDifferentThread(null, null, 0, eventName + "_uploadcompleted", false, new Object[] { Url, RemoteFile, true});
				}
				catch (Exception e) {
					ba.setLastException(e);
					ba.raiseEventFromDifferentThread(null, null, 0, eventName + "_uploadcompleted", false, new Object[] { Url, RemoteFile, false});
				}
			}
		};
		ba.submitRunnable(r, this, 0);
	}

	/**
	 *Downloads a remote file and stores it in the specified path.
	 *The DownloadCompleted event will be raised when the process completes.
	 *Example: <code>SMB1.DownloadFile("smb://USER-PC/Users/Public/", "1.jpg", File.DirRootExternal, "1.jpg")</code>
	 */
	public void DownloadFile(final BA ba, final String Url, final String RemoteFile, String LocalDir, String LocalFile) throws FileNotFoundException {
		OutputStreamWrapper o = new OutputStreamWrapper();
		o = File.OpenOutput(LocalDir, LocalFile, false);
		DownloadFile2(ba, Url, RemoteFile, o.getObject(), true);
	}

	/**
	 *Similar to DownloadFile. Expects an OutputStream instead of the local file.
	 *If Close is True then the OutputStream is closed on completion.
	 */
	public void DownloadFile2(final BA ba, final String Url, final String RemoteFile, final OutputStream out, final boolean Close) {
		Runnable r = new Runnable() {
			@Override
			public void run() {
				try {
					SmbFile sf = createSmb(Url, RemoteFile);
					InputStream in = sf.getInputStream();
					File.Copy2(in, out);
					out.flush();
					if (Close)
						out.close();
					ba.raiseEventFromDifferentThread(null, null, 0, eventName + "_downloadcompleted", false, new Object[] { Url, RemoteFile, true});
				}
				catch (Exception e) {
					ba.setLastException(e);
					ba.raiseEventFromDifferentThread(null, null, 0, eventName + "_downloadcompleted", false, new Object[] { Url, RemoteFile, false});
				}
			}
		};
		ba.submitRunnable(r, this, 0);
	}

	/**
	 *Lists all files and folders in the given remote path.
	 *FilterPattern - DOS filter pattern. Pass an empty string to list all files and folders.
	 *The ListCompleted event will be raised when the process completes.
	 *Example: <code>SMB1.ListFiles("smb://USER-PC/Users/Public/", "*.jpg")</code>
	 */
	public void ListFiles(final BA ba, final String Url, final String FilterPattern) {
		Runnable r = new Runnable() {
			@Override
			public void run() {
				SmbFileWrapper[] files;
				try {
					SmbFile f = createSmb(Url, "");
					SmbFile[] ff;
					if (FilterPattern.length() > 0)
						ff = f.listFiles(FilterPattern);
					else
						ff = f.listFiles();
					files = new SmbFileWrapper[ff.length];
					for (int i = 0;i < ff.length;i++) {
						files[i] = (SmbFileWrapper)AbsObjectWrapper.ConvertToWrapper(new SmbFileWrapper(), ff[i]);
					}
					ba.raiseEventFromDifferentThread(null, null, 0, eventName + "_listcompleted", false, new Object[] { Url, true, files});
				}
				catch (Exception e) {
					ba.setLastException(e);
					files = new SmbFileWrapper[0];
					ba.raiseEventFromDifferentThread(null, null, 0, eventName + "_listcompleted", false, new Object[] { Url, false, files});
				}
			}
		};
		ba.submitRunnable(r, this, 0);
	}

	// ADDITIONS V1.1

	 //mkDirs(), renameTo(), canRead(), canWrite(), and getFreeMemory()? 
	
	/**
	 *This method returns the free disk space in bytes of the drive on which the directory or file resides.
	 *The FreeSpaceCompleted event will be raised when the process completes. 
	 *Example: <code>SMB1.GetDiskFreeSpace("smb://USER-PC/", "")</code>
	 */
	public void GetDiskFreeSpace(final BA ba, final String Url, final String RemoteFile) {
		Runnable r = new Runnable() {
			@Override
			public void run() {
				try {
					SmbFile sf = createSmb(Url, RemoteFile);
					long size = sf.getDiskFreeSpace();
					ba.raiseEventFromDifferentThread(null, null, 0, eventName + "_freespacecompleted", false, new Object[] { Url, RemoteFile, size, true});
				}
				catch (Exception e) {
					ba.setLastException(e);
					ba.raiseEventFromDifferentThread(null, null, 0, eventName + "_freespacecompleted", false, new Object[] { Url, RemoteFile, 0, false});
				}
			}
		};
		ba.submitRunnable(r, this, 0);
	}
	
	/**
	 *This method Creates a directory with the path specified by and creating any parent directories that do not exist.
	 *The MakeDirCompleted event will be raised when the process completes. 
	 *Example: <code>SMB1.MakeDir("smb://USER-PC/Users/Public/", "NewDir")</code>
	 */
	public void MakeDir(final BA ba, final String Url, final String RemoteFile) {
		Runnable r = new Runnable() {
			@Override
			public void run() {
				try {
					SmbFile sf = createSmb(Url, RemoteFile);
					sf.mkdirs();
					ba.raiseEventFromDifferentThread(null, null, 0, eventName + "_makedircompleted", false, new Object[] { Url, RemoteFile, true});
				}
				catch (Exception e) {
					ba.setLastException(e);
					ba.raiseEventFromDifferentThread(null, null, 0, eventName + "_makedircompleted", false, new Object[] { Url, RemoteFile, false});
				}
			}
		};
		ba.submitRunnable(r, this, 0);
	}
	
	/**
	 *This method will copy the source file or directory specified to the destination specified.
	 *The CopyCompleted event will be raised when the process completes.
	 *Example: <code>SMB1.Copy("smb://USER-PC/", "1.jpg", "smb://USER-PC/", "2.jpg")</code>
	 */
	public void CopyFile(final BA ba, final String SrcUrl, final String SrcRemoteFile, final String DestUrl, final String DestRemoteFile) {
		Runnable r = new Runnable() {
			@Override
			public void run() {
				try {
					SmbFile srcsf = createSmb(SrcUrl, SrcRemoteFile);
					SmbFile destsf = createSmb(DestUrl, DestRemoteFile);
					srcsf.copyTo(destsf);
					ba.raiseEventFromDifferentThread(null, null, 0, eventName + "_copycompleted", false, new Object[] { DestUrl, DestRemoteFile, true});
				}
				catch (Exception e) {
					ba.setLastException(e);
					ba.raiseEventFromDifferentThread(null, null, 0, eventName + "_copycompleted", false, new Object[] { DestUrl, DestRemoteFile, false});
				}
			}
		};
		ba.submitRunnable(r, this, 0);
	}

	/**
	 *This method will rename the file or directory specified.
	 *The RenameCompleted event will be raised when the process completes.
	 *Example: <code>SMB1.Rename("smb://USER-PC/", "1.jpg", "2.jpg)</code>
	 */
	public void RenameFile(final BA ba, final String Url, final String SrcRemoteFile, final String DestRemoteFile) {
		Runnable r = new Runnable() {
			@Override
			public void run() {
				try {
					SmbFile srcsf = createSmb(Url, SrcRemoteFile);
					SmbFile destsf = createSmb(Url, DestRemoteFile);
					srcsf.renameTo(destsf);
					ba.raiseEventFromDifferentThread(null, null, 0, eventName + "_renamecompleted", false, new Object[] { Url, DestRemoteFile, true});
				}
				catch (Exception e) {
					ba.setLastException(e);
					ba.raiseEventFromDifferentThread(null, null, 0, eventName + "_renamecompleted", false, new Object[] { Url, DestRemoteFile, false});
				}
			}
		};
		ba.submitRunnable(r, this, 0);
	}	
	
	
	
	
	//###########################################################################
	// SmbFileWrapper
	//###########################################################################
	
	
	/**
	 *Stores the metadata for a file or folder entry retrieved by calling SMB.ListFiles. 
	 */
	@ShortName("SMBFile")
	public static class SmbFileWrapper extends AbsObjectWrapper<SmbFile> {

		/**
		 *Gets the file name.
		 */
		public String getName() {
			return getObject().getName();
		}

		/**
		 *Returns true if this entry is a directory.
		 */
		public boolean getDirectory() throws SmbException {
			return getObject().isDirectory();
		}

		/**
		 *Returns the last modified date.
		 *Example: <code>DateTime.Date(File.LastModified)</code>
		 */
		public long getLastModified() {
			return getObject().getLastModified();
		}

		/**
		 *Returns the file size.
		 */
		public long getSize() throws SmbException {
			return getObject().length();
		}

		/**
		 *Returns the Url of the file parent.
		 */
		public String getParent() {
			return getObject().getParent();
		}

		
		// ADDITIONS V1.1
		
		/**
		 *Returns True if the file can be read.
		 */
		public boolean CanRead() throws SmbException
		{
			return getObject().canRead();
		}
		
		/**
		 *Returns True if the file is not marked read-only.
		 */
		public boolean CanWrite() throws SmbException
		{
			return getObject().canWrite();
		}
		
		/**
		 *Returns True if the file is marked as hidden
		 */
		public boolean IsHidden() throws SmbException
		{
			return getObject().isHidden();
		}
		
		
	
	}
}
